//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>

#include "definitions.h"
#include "f11asrt.h"

void f11_error(const char* msg, ...)
{
    va_list args;

    va_start (args, msg);
    fprintf(stderr, "%%FTU-E-ERROR: ");
    vfprintf(stderr, msg, args);
    fprintf(stderr, "\n");
    fflush(stderr);

    va_end (args);
}

void f11_info(const char* msg, ...)
{
    va_list args;

    va_start (args, msg);
    fprintf(stderr, "%%FTU-I-INFO: ");
    vfprintf(stderr, msg, args);
    fprintf(stderr, "\n");
    fflush(stderr);

    va_end (args);
}

#if defined(_DEBUG)

int f11_trace_enabled = 0;
void f11_trace(const char* msg, ...) 
{
    if (f11_trace_enabled) {
        va_list args;
        va_start (args, msg);
	fprintf(stderr, "-->");
        vfprintf(stderr, msg, args);
	fflush(stderr);
        va_end (args);
    }	
}

void f11_enable_trace()
{
    f11_trace_enabled = 1;
}

void f11_disable_trace()
{
    f11_trace_enabled = 0;
}

#if defined(unix) && !defined(sun)
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <sys/stat.h>
#include <stropts.h>
#include <string.h>

int __assert_pid = 0;

#define ASSERT_LOCK ".debug-assert-lock"
#define ASSERT_EL   ".debug-assert.el"
#define ASSERT_DBG  ".debug-assert"

static char assert_lock[FILENAME_MAX + PATH_MAX + 1]	= {0};
static char assert_el[FILENAME_MAX + PATH_MAX + 1]	= {0};
static char assert_dbg[FILENAME_MAX + PATH_MAX + 1] 	= {0};
static char assert_image[FILENAME_MAX + PATH_MAX + 1] 	= {0};

static char cwd[PATH_MAX + 1];

//  OSF specific
int __under_debug()
{
    char 		buffer[64];
    int	 		proc, ret;
    struct prstatus 	stat;

    sprintf(buffer, "/proc/%d", getpid());
    proc = open(buffer, O_RDONLY, 0666);
    if(proc == -1) return -1;

    if(ioctl(proc, PIOCSTATUS, &stat) == -1) {
	close(proc);
	return -1;
    }

    ret = stat.pr_flags & (PR_PTRACE|PR_TRACING) ? 1 : 0;
    close(proc);
    return ret;
}

// end of OSF specific
static void wrote_debug_assert( char* fileName, int line )
{
    FILE*   f;

    sprintf(assert_dbg, "%s/%s%d", getwd(cwd), ASSERT_DBG, getpid());
    f = fopen(assert_dbg, "w+");
    if(!f) {
	perror("Can't create/open dgb command file. ");
	return;
    }
    fprintf(f, "b %s:%d\n", fileName, line);
    fprintf(f, "continue\nfinish\nclear %s:%d\n", fileName, line);
    fclose(f);
}

static void wrote_emacs_startup()
{
    FILE    *f, *f1;

    sprintf(assert_el, "%s/%s%d", getwd(cwd), ASSERT_EL, getpid());
    sprintf(assert_lock, "%s/%s%d", getwd(cwd), ASSERT_LOCK, getpid());

    f = fopen(assert_el, "w+");
    if(!f) {
	perror("Can't create/open emacs (assert-debug) startup file.");
	return;
    }

    f1 = fopen(assert_lock, "w+");
    if(!f) {
	perror("Can't create debug-assert lock file.");
	return;
    }

    fclose(f1);
    
    fprintf(f, "(gdb \"/usr/local/bin/gdb -q -s %s -x"\
	    " %s %s %d\")\n", assert_image, assert_dbg,
	    assert_image, getpid());

    fprintf(f, "(delete-file \"%s\")\n", assert_lock);
    fclose(f);
}

void __do_assert( char* fileName, int line )
{
    char choise;
    int	 j;
    char buffer[ 64 ];

    fprintf( stderr, "(ignore,debug,abort) (i/d/a) ? " );

    read( 1, &choise, 1 );
    fprintf( stderr, "\n" );

    switch( choise ) {
      case 'd':
	{
	    int i = 2;
	    while( i-- ) {
		if( !i ) {
		    fprintf( stderr, "\n***** Loading ... *****\n" );
		    j = 20;
		    while(j--) {
			struct stat s;
			sleep(1);
			if( stat(assert_lock, &s) == -1 )
			    break;
		    }

		    // Waiting for debugger loading
		    sleep(2);

		    remove(assert_el);
		    remove(assert_dbg);

		}
		line = __LINE__;
		if( !i )  return;

		sprintf( buffer, "%d", getpid() );
		wrote_debug_assert( __FILE__, line );
		wrote_emacs_startup();
		if( !(__assert_pid = fork()) ){
		    execl("/usr/local/bin/emacs", 
			  "/usr/local/bin/emacs", 
			  "-l", assert_el, 
			  NULL);
		    perror( "Can't execute '/usr/local/bin/emacs'." );
		    exit(10);
		}
	    }
	    break;
	}
      case 'a':
	exit(-2);
      case 'i':
      default:
	return;
    }
    return;
}

#undef main(a,b)

extern int user_main(int argc, char* argv[]);

void init_assert(int argc, char* argv[])
{
    if(argv[0][0] == '/') {
	strcpy(assert_image, argv[0]);
    } else {
	sprintf(assert_image, "%s/%s", getwd(cwd), argv[0]);
    }
}

int main(int argc, char* argv[])
{
    init_assert(argc, argv);
    return user_main(argc, argv);
}

#endif // UNIX
#endif // _DEBUG
